//
//  Copyright (C) 2013-2024 Paolo Tosco and other RDKit contributors
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
#include <RDGeneral/export.h>
#ifndef RD_UFFDISTANCECONSTRAINT_H
#define RD_UFFDISTANCECONSTRAINT_H
#include <ForceField/DistanceConstraint.h>

namespace ForceFields {
namespace UFF {
using DistanceConstraintContrib = ForceFields::DistanceConstraintContrib;
}  // namespace UFF
}  // namespace ForceFields
#endif
