#line 2 "/Users/landrgr1/RDKit_git/Code/GraphMol/SLNParse/lex.yysln.cpp"

#line 4 "/Users/landrgr1/RDKit_git/Code/GraphMol/SLNParse/lex.yysln.cpp"

#define  YY_INT_ALIGNED short int

/* A lexical scanner generated by flex */

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 5
#define YY_FLEX_SUBMINOR_VERSION 35
#if YY_FLEX_SUBMINOR_VERSION > 0
#define FLEX_BETA
#endif

/* First, we deal with  platform-specific or compiler-specific issues. */

/* begin standard C headers. */
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <cstdlib>

/* end standard C headers. */

/* flex integer type definitions */

#ifndef FLEXINT_H
#define FLEXINT_H

/* C99 systems have <inttypes.h>. Non-C99 systems may or may not. */

#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L

/* C99 says to define __STDC_LIMIT_MACROS before including stdint.h,
 * if you want the limit (max/min) macros for int types. 
 */
#ifndef __STDC_LIMIT_MACROS
#define __STDC_LIMIT_MACROS 1
#endif

#include <cinttypes>
typedef int8_t flex_int8_t;
typedef uint8_t flex_uint8_t;
typedef int16_t flex_int16_t;
typedef uint16_t flex_uint16_t;
typedef int32_t flex_int32_t;
typedef uint32_t flex_uint32_t;
typedef uint64_t flex_uint64_t;
#else
typedef signed char flex_int8_t;
typedef short int flex_int16_t;
typedef int flex_int32_t;
typedef unsigned char flex_uint8_t; 
typedef unsigned short int flex_uint16_t;
typedef unsigned int flex_uint32_t;
#endif /* ! C99 */

/* Limits of integral types. */
#ifndef INT8_MIN
#define INT8_MIN               (-128)
#endif
#ifndef INT16_MIN
#define INT16_MIN              (-32767-1)
#endif
#ifndef INT32_MIN
#define INT32_MIN              (-2147483647-1)
#endif
#ifndef INT8_MAX
#define INT8_MAX               (127)
#endif
#ifndef INT16_MAX
#define INT16_MAX              (32767)
#endif
#ifndef INT32_MAX
#define INT32_MAX              (2147483647)
#endif
#ifndef UINT8_MAX
#define UINT8_MAX              (255U)
#endif
#ifndef UINT16_MAX
#define UINT16_MAX             (65535U)
#endif
#ifndef UINT32_MAX
#define UINT32_MAX             (4294967295U)
#endif

#endif /* ! FLEXINT_H */

#ifdef __cplusplus

/* The "const" storage-class-modifier is valid. */
#define YY_USE_CONST

#else	/* ! __cplusplus */

/* C99 requires __STDC__ to be defined as 1. */
#if defined (__STDC__)

#define YY_USE_CONST

#endif	/* defined (__STDC__) */
#endif	/* ! __cplusplus */

#ifdef YY_USE_CONST
#define yyconst const
#else
#define yyconst
#endif

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an unsigned
 * integer for use as an array index.  If the signed char is negative,
 * we want to instead treat it as an 8-bit unsigned char, hence the
 * double cast.
 */
#define YY_SC_TO_UI(c) ((unsigned int) (unsigned char) c)

/* An opaque pointer. */
#ifndef YY_TYPEDEF_YY_SCANNER_T
#define YY_TYPEDEF_YY_SCANNER_T
typedef void* yyscan_t;
#endif

/* For convenience, these vars (plus the bison vars far below)
   are macros in the reentrant scanner. */
#define yyin yyg->yyin_r
#define yyout yyg->yyout_r
#define yyextra yyg->yyextra_r
#define yyleng yyg->yyleng_r
#define yytext yyg->yytext_r
#define yylineno (YY_CURRENT_BUFFER_LVALUE->yy_bs_lineno)
#define yycolumn (YY_CURRENT_BUFFER_LVALUE->yy_bs_column)
#define yy_flex_debug yyg->yy_flex_debug_r

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN yyg->yy_start = 1 + 2 *

/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START ((yyg->yy_start - 1) / 2)
#define YYSTATE YY_START

/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)

/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE yysln_restart(yyin ,yyscanner )

#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#ifndef YY_BUF_SIZE
#define YY_BUF_SIZE 16384
#endif

/* The state buf must be large enough to hold one state per character in the main buffer.
 */
#define YY_STATE_BUF_SIZE   ((YY_BUF_SIZE + 2) * sizeof(yy_state_type))

#ifndef YY_TYPEDEF_YY_BUFFER_STATE
#define YY_TYPEDEF_YY_BUFFER_STATE
typedef struct yy_buffer_state *YY_BUFFER_STATE;
#endif

#ifndef YY_TYPEDEF_YY_SIZE_T
#define YY_TYPEDEF_YY_SIZE_T
typedef size_t yy_size_t;
#endif

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2

    #define YY_LESS_LINENO(n)
    
/* Return all but the first "n" matched characters back to the input stream. */
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		*yy_cp = yyg->yy_hold_char; \
		YY_RESTORE_YY_MORE_OFFSET \
		yyg->yy_c_buf_p = yy_cp = yy_bp + yyless_macro_arg - YY_MORE_ADJ; \
		YY_DO_BEFORE_ACTION; /* set up yytext again */ \
		} \
	while ( 0 )

#define unput(c) yyunput( c, yyg->yytext_ptr , yyscanner )

#ifndef YY_STRUCT_YY_BUFFER_STATE
#define YY_STRUCT_YY_BUFFER_STATE
struct yy_buffer_state
	{
	FILE *yy_input_file;

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	yy_size_t yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	yy_size_t yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

    int yy_bs_lineno; /**< The line count. */
    int yy_bs_column; /**< The column count. */
    
	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;

#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via yysln_restart()), so that the user can continue scanning by
	 * just pointing yyin at a new input file.
	 */
#define YY_BUFFER_EOF_PENDING 2

	};
#endif /* !YY_STRUCT_YY_BUFFER_STATE */

/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 *
 * Returns the top of the stack, or NULL.
 */
#define YY_CURRENT_BUFFER ( yyg->yy_buffer_stack \
                          ? yyg->yy_buffer_stack[yyg->yy_buffer_stack_top] \
                          : NULL)

/* Same as previous macro, but useful when we know that the buffer stack is not
 * NULL or when we need an lvalue. For internal use only.
 */
#define YY_CURRENT_BUFFER_LVALUE yyg->yy_buffer_stack[yyg->yy_buffer_stack_top]

void yysln_restart (FILE *input_file ,yyscan_t yyscanner );
void yysln__switch_to_buffer (YY_BUFFER_STATE new_buffer ,yyscan_t yyscanner );
YY_BUFFER_STATE yysln__create_buffer (FILE *file,int size ,yyscan_t yyscanner );
void yysln__delete_buffer (YY_BUFFER_STATE b ,yyscan_t yyscanner );
void yysln__flush_buffer (YY_BUFFER_STATE b ,yyscan_t yyscanner );
void yysln_push_buffer_state (YY_BUFFER_STATE new_buffer ,yyscan_t yyscanner );
void yysln_pop_buffer_state (yyscan_t yyscanner );

static void yysln_ensure_buffer_stack (yyscan_t yyscanner );
static void yysln__load_buffer_state (yyscan_t yyscanner );
static void yysln__init_buffer (YY_BUFFER_STATE b,FILE *file ,yyscan_t yyscanner );

#define YY_FLUSH_BUFFER yysln__flush_buffer(YY_CURRENT_BUFFER ,yyscanner)

YY_BUFFER_STATE yysln__scan_buffer (char *base,yy_size_t size ,yyscan_t yyscanner );
YY_BUFFER_STATE yysln__scan_string (yyconst char *yy_str ,yyscan_t yyscanner );
YY_BUFFER_STATE yysln__scan_bytes (yyconst char *bytes,yy_size_t len ,yyscan_t yyscanner );

void *yysln_alloc (yy_size_t ,yyscan_t yyscanner );
void *yysln_realloc (void *,yy_size_t ,yyscan_t yyscanner );
void yysln_free (void * ,yyscan_t yyscanner );

#define yy_new_buffer yysln__create_buffer

#define yy_set_interactive(is_interactive) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){ \
        yysln_ensure_buffer_stack (yyscanner); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yysln__create_buffer(yyin,YY_BUF_SIZE ,yyscanner); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_is_interactive = is_interactive; \
	}

#define yy_set_bol(at_bol) \
	{ \
	if ( ! YY_CURRENT_BUFFER ){\
        yysln_ensure_buffer_stack (yyscanner); \
		YY_CURRENT_BUFFER_LVALUE =    \
            yysln__create_buffer(yyin,YY_BUF_SIZE ,yyscanner); \
	} \
	YY_CURRENT_BUFFER_LVALUE->yy_at_bol = at_bol; \
	}

#define YY_AT_BOL() (YY_CURRENT_BUFFER_LVALUE->yy_at_bol)

/* Begin user sect3 */

#define yysln_wrap(n) 1
#define YY_SKIP_YYWRAP

typedef unsigned char YY_CHAR;

typedef int yy_state_type;

#define yytext_ptr yytext_r

static yy_state_type yy_get_previous_state (yyscan_t yyscanner );
static yy_state_type yy_try_NUL_trans (yy_state_type current_state  ,yyscan_t yyscanner);
static int yy_get_next_buffer (yyscan_t yyscanner );
static void yy_fatal_error (yyconst char msg[] ,yyscan_t yyscanner );

/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION \
	yyg->yytext_ptr = yy_bp; \
	yyleng = (yy_size_t) (yy_cp - yy_bp); \
	yyg->yy_hold_char = *yy_cp; \
	*yy_cp = '\0'; \
	yyg->yy_c_buf_p = yy_cp;

#define YY_NUM_RULES 170
#define YY_END_OF_BUFFER 171
/* This struct is not used in this scanner,
   but its presence is necessary. */
struct yy_trans_info
	{
	flex_int32_t yy_verify;
	flex_int32_t yy_nxt;
	};
static yyconst flex_int16_t yy_accept[249] =
    {   0,
        0,    0,    0,    0,    0,    0,    0,    0,    1,    1,
        0,    0,  171,  169,  168,  154,  128,  153,  131,  132,
      166,  127,  126,  163,  167,  130,  152,  159,  142,  165,
      169,   97,   98,  169,  169,  103,  169,  113,  106,   14,
      169,  169,   99,  100,  101,  169,  102,  169,   85,   18,
       67,  169,   33,  169,  155,  129,  154,  138,  136,  137,
        2,    2,    2,  158,    2,    2,  146,  130,  145,  139,
      160,    4,  169,    3,  169,  148,    3,    3,    3,    3,
        3,  147,  161,    1,    1,    1,  150,    1,    1,    1,
        1,    1,    1,    1,    1,  149,    1,  162,    1,    1,

      157,  151,    1,  144,  164,  143,  156,  167,   82,   41,
       11,   88,    0,   13,   28,   78,   72,   49,    7,   76,
       90,  105,   15,   42,   51,   91,  104,   89,   21,   19,
       48,   24,   59,   61,   92,   56,   22,   93,   80,   26,
       57,   27,  112,  111,    0,    5,   65,   73,   60,   43,
       70,   30,   50,    6,   96,   64,   94,   10,   20,   36,
        9,   35,   53,    8,   23,   95,   86,   69,   84,   75,
       40,   54,   77,   52,   71,   87,   81,   31,   68,   39,
       79,   38,   45,   16,   29,   12,   55,   44,   32,   66,
       58,   37,   46,   83,   17,   74,   62,   47,   63,   25,

       34,  133,  135,  134,    2,    2,    2,    2,    2,    2,
        2,    2,    2,    2,  140,    4,    4,  141,    3,    3,
        3,    1,    1,  107,  109,  110,  108,  117,  115,    2,
        2,    2,    2,  116,  114,    2,    2,    2,    2,  125,
      122,  123,  119,  124,  120,  121,  118,    0
    } ;

static yyconst flex_int32_t yy_ec[256] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    2,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    3,    4,    5,    6,    1,    1,    7,    1,    8,
        9,   10,   11,   12,   13,   14,    1,   15,   15,   15,
       15,   15,   15,   15,   15,   15,   15,   16,   17,   18,
       19,   20,    1,   21,   22,   23,   24,   25,   26,   27,
       28,   29,   30,   31,   32,   33,   34,   35,   36,   37,
       31,   38,   39,   40,   41,   42,   43,   44,   45,   46,
       47,    1,   48,   49,   50,    1,   51,   52,   53,   54,

       55,   56,   57,   58,   59,   31,   60,   61,   62,   63,
       64,   65,   31,   66,   67,   68,   69,   70,   31,   31,
       71,   31,    1,   72,    1,   73,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,

        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1
    } ;

static yyconst flex_int32_t yy_meta[74] =
    {   0,
        1,    1,    2,    3,    2,    1,    3,    2,    2,    1,
        1,    4,    4,    5,    6,    1,    3,    1,    1,    3,
        1,    6,    6,    6,    6,    6,    6,    6,    6,    6,
        6,    6,    6,    6,    6,    6,    6,    6,    6,    6,
        6,    6,    6,    6,    6,    6,    1,    3,    1,    6,
        6,    6,    6,    6,    6,    6,    6,    6,    6,    6,
        6,    6,    6,    6,    6,    6,    6,    6,    6,    6,
        6,    3,    1
    } ;

static yyconst flex_int16_t yy_base[258] =
    {   0,
        0,    0,   73,    0,  146,  219,  292,  365,  437,  509,
      161,  162,  329,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  313,  702,  702,  702,  702,  702,
      119,  146,  190,  256,  123,  136,  149,  267,  145,  260,
      148,  201,  219,  254,  281,  336,  345,  300,  702,  702,
      702,  265,  267,  150,  702,  702,  297,  285,  702,  276,
      235,  410,  406,  702,  454,  450,  702,  270,  702,  702,
      702,  265,  267,  271,  270,  702,  258,  256,  255,  244,
      238,  702,  702,    0,    0,    0,  702,    0,    0,    0,
        0,    0,    0,  227,    0,  702,    0,  702,    0,    0,

      702,  702,    0,  702,  702,  702,  702,  200,  702,  702,
      702,  702,  113,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  122,  107,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  702,  702,

      702,  702,  702,  702,    0,  275,  377,  421,  493,  519,
      523,  533,  538,  550,  702,    0,  267,  702,  166,  702,
      257,    0,  155,  702,  702,  702,  702,  702,  702,  555,
      579,  584,  589,  702,  702,  597,  602,  607,  612,  702,
      702,  702,  702,  702,  702,  702,  702,  702,  662,  668,
      674,  677,  679,  684,  690,  693,  695
    } ;

static yyconst flex_int16_t yy_def[258] =
    {   0,
      248,    1,  248,    3,  249,  249,  250,  250,  251,  251,
        1,    1,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      252,  252,  252,  248,  252,  252,  248,  248,  248,  248,
      248,  253,  248,  254,  254,  248,  254,  254,  254,  254,
      254,  248,  248,  255,  255,  255,  248,  255,  255,  255,
      255,  255,  255,  255,  255,  248,  255,  248,  255,  255,

      248,  248,  255,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,

      248,  248,  248,  248,  256,  252,  252,  252,  252,  252,
      252,  252,  252,  252,  248,  257,  253,  248,  254,  248,
      254,  255,  255,  248,  248,  248,  248,  248,  248,  252,
      252,  252,  252,  248,  248,  252,  252,  252,  252,  248,
      248,  248,  248,  248,  248,  248,  248,    0,  248,  248,
      248,  248,  248,  248,  248,  248,  248
    } ;

static yyconst flex_int16_t yy_nxt[776] =
    {   0,
       14,   15,   14,   16,   14,   17,   18,   19,   20,   21,
       22,   14,   23,   24,   25,   26,   27,   28,   29,   14,
       30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
       14,   40,   41,   42,   43,   44,   45,   46,   47,   48,
       49,   50,   51,   52,   53,   54,   55,   14,   14,   14,
       14,   14,   14,   14,   14,   14,   14,   14,   14,   14,
       14,   14,   14,   14,   14,   14,   14,   14,   14,   14,
       14,   14,   56,   14,   15,   14,   57,   14,   17,   18,
       19,   20,   21,   22,   14,   23,   24,   25,   26,   27,
       58,   59,   60,   30,   61,   61,   61,   61,   61,   61,

       61,   61,   62,   61,   61,   61,   61,   63,   61,   61,
       61,   61,   61,   61,   61,   61,   61,   61,   61,   55,
       64,   14,   14,   61,   61,   61,   61,   61,   61,   61,
       61,   65,   61,   61,   61,   66,   61,   61,   61,   61,
       61,   61,   61,   61,   14,   56,   14,   15,   14,   16,
       14,   17,   67,   19,   20,   21,   22,   14,   23,   24,
       25,   68,   69,   28,   70,   71,   30,  104,  104,  223,
      220,  109,  105,  105,  226,  110,  227,  106,  106,  111,
      112,  113,  225,  224,  114,  115,  116,  117,  134,  135,
      137,  136,   55,   14,   73,   14,  118,  138,  153,  140,

      119,  139,  141,  142,  120,  121,  154,  150,  107,  107,
      151,  122,  200,  155,  108,  201,  156,   14,   56,   14,
       15,   14,   16,   14,   17,   67,   19,   20,   21,   22,
       14,   23,   24,   25,   68,   69,   28,   70,   71,   30,
      123,  223,  220,  124,  125,  126,  205,  205,  220,  205,
      127,  128,  221,  129,  157,  130,  131,  158,  132,  220,
      220,  220,  220,  159,  160,   55,   14,   73,   14,  161,
      162,  221,  163,  164,  248,  220,  143,  165,  216,  216,
      216,  216,  166,  167,  205,  218,  205,  205,  215,  205,
       14,   56,   14,   15,  204,   16,   75,   17,   76,   77,

       78,   21,   22,  203,   79,   80,   81,   26,   82,   28,
       14,   83,   30,  144,  216,  202,  216,  145,  199,  198,
      168,  146,  147,  148,  205,  152,  133,  108,  248,  248,
      149,  169,  170,  248,  171,  248,  248,  248,   55,   14,
       14,  248,  172,  248,  173,  248,  174,  248,  175,  176,
      190,  191,  192,  248,  193,  248,  248,  194,  195,  248,
      196,  197,  248,   14,   56,   14,   15,  248,   16,   75,
       17,   76,   77,   78,   21,   22,  248,   79,   80,   81,
       26,   82,   28,   14,   83,   30,  177,  178,  205,  205,
      179,  205,  248,  180,  248,  228,  183,  184,  181,  185,

      248,  248,  248,  186,  182,  248,  187,  188,  248,  248,
      189,   55,   14,   14,  248,  248,  248,  205,  205,  248,
      205,  205,  205,  248,  205,  248,  205,  248,  248,  248,
      248,  248,  205,  205,  248,  205,   14,   56,   85,  229,
       16,  209,   86,   87,   88,   89,   90,   91,  207,   92,
       93,   94,   95,   96,   97,  205,   98,   99,  248,  205,
      248,  205,  205,  248,  205,  205,  205,  248,  205,  210,
      205,  248,  248,  248,  248,  248,  208,  248,  248,  248,
      248,  248,  248,  100,  101,  213,  248,  248,  248,  248,
      248,  248,  211,  248,  248,  248,  248,  248,  248,  205,

      248,  248,  248,  205,  205,  205,  248,  205,  102,  103,
       85,  248,   16,  214,   86,   87,   88,   89,   90,   91,
      212,   92,   93,   94,   95,   96,   97,  248,   98,   99,
      205,  205,  230,  205,  205,  205,  248,  205,  248,  248,
      248,  234,  205,  248,  205,  205,  248,  205,  248,  205,
      205,  235,  205,  248,  248,  100,  101,  248,  232,  248,
      231,  205,  205,  248,  205,  248,  205,  205,  205,  205,
      248,  248,  205,  240,  248,  248,  248,  236,  248,  248,
      102,  103,  205,  248,  248,  248,  233,  205,  248,  238,
      205,  205,  248,  205,  248,  205,  205,  241,  205,  205,

      205,  205,  242,  205,  205,  237,  248,  243,  205,  205,
      248,  205,  248,  205,  205,  244,  205,  239,  205,  205,
      245,  205,  248,  205,  205,  246,  205,  248,  205,  248,
      247,  248,  248,  205,  248,  248,  248,  248,  205,  248,
      248,  248,  248,  248,  248,  248,  205,  248,  248,  248,
      248,  205,  248,  248,  248,  248,  205,  248,  248,  248,
      248,  205,   72,   72,   72,   72,   72,   72,   74,   74,
       74,   74,   74,   74,   84,   84,   84,   84,   84,   84,
      206,  248,  206,  217,  217,  219,  248,  219,  219,  219,
      222,  222,  248,  222,  222,  222,  205,  248,  205,  216,

      216,   13,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248
    } ;

static yyconst flex_int16_t yy_chk[776] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,

        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
        3,    3,    3,    3,    3,    3,    5,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,    5,    5,    5,
        5,    5,    5,    5,    5,    5,    5,   11,   12,  223,
      219,   31,   11,   12,  146,   31,  146,   11,   12,   31,
       31,   31,  145,  113,   31,   31,   31,   31,   35,   35,
       36,   35,    5,    5,    5,    5,   32,   36,   41,   37,

       32,   36,   37,   37,   32,   32,   41,   39,   11,   12,
       39,   32,   54,   41,  108,   54,   41,    5,    5,    6,
        6,    6,    6,    6,    6,    6,    6,    6,    6,    6,
        6,    6,    6,    6,    6,    6,    6,    6,    6,    6,
       33,   94,   81,   33,   33,   33,   61,   61,   80,   61,
       33,   33,   81,   33,   42,   33,   33,   42,   33,   79,
       78,  221,   77,   42,   42,    6,    6,    6,    6,   43,
       43,  221,   43,   43,   75,   74,   38,   43,   72,   72,
      217,  217,   43,   43,   61,   73,  206,  206,   68,  206,
        6,    6,    7,    7,   60,    7,    7,    7,    7,    7,

        7,    7,    7,   58,    7,    7,    7,    7,    7,    7,
        7,    7,    7,   38,   72,   57,  217,   38,   53,   52,
       44,   38,   38,   38,  206,   40,   34,   25,   13,    0,
       38,   45,   45,    0,   45,    0,    0,    0,    7,    7,
        7,    0,   45,    0,   45,    0,   45,    0,   45,   45,
       48,   48,   48,    0,   48,    0,    0,   48,   48,    0,
       48,   48,    0,    7,    7,    8,    8,    0,    8,    8,
        8,    8,    8,    8,    8,    8,    0,    8,    8,    8,
        8,    8,    8,    8,    8,    8,   46,   46,  207,  207,
       46,  207,    0,   46,    0,  207,   47,   47,   46,   47,

        0,    0,    0,   47,   46,    0,   47,   47,    0,    0,
       47,    8,    8,    8,    0,    0,    0,   63,   63,    0,
       63,   62,   62,    0,   62,    0,  207,    0,    0,    0,
        0,    0,  208,  208,    0,  208,    8,    8,    9,  208,
        9,   63,    9,    9,    9,    9,    9,    9,   62,    9,
        9,    9,    9,    9,    9,   63,    9,    9,    0,   62,
        0,   66,   66,    0,   66,   65,   65,    0,   65,   63,
      208,    0,    0,    0,    0,    0,   62,    0,    0,    0,
        0,    0,    0,    9,    9,   66,    0,    0,    0,    0,
        0,    0,   65,    0,    0,    0,    0,    0,    0,   66,

        0,    0,    0,   65,  209,  209,    0,  209,    9,    9,
       10,    0,   10,   66,   10,   10,   10,   10,   10,   10,
       65,   10,   10,   10,   10,   10,   10,    0,   10,   10,
      210,  210,  209,  210,  211,  211,    0,  211,    0,    0,
        0,  211,  209,    0,  212,  212,    0,  212,    0,  213,
      213,  212,  213,    0,    0,   10,   10,    0,  210,    0,
      209,  214,  214,    0,  214,    0,  230,  230,  210,  230,
        0,    0,  211,  230,    0,    0,    0,  213,    0,    0,
       10,   10,  212,    0,    0,    0,  210,  213,    0,  214,
      231,  231,    0,  231,    0,  232,  232,  231,  232,  214,

      233,  233,  232,  233,  230,  213,    0,  233,  236,  236,
        0,  236,    0,  237,  237,  236,  237,  214,  238,  238,
      237,  238,    0,  239,  239,  238,  239,    0,  231,    0,
      239,    0,    0,  232,    0,    0,    0,    0,  233,    0,
        0,    0,    0,    0,    0,    0,  236,    0,    0,    0,
        0,  237,    0,    0,    0,    0,  238,    0,    0,    0,
        0,  239,  249,  249,  249,  249,  249,  249,  250,  250,
      250,  250,  250,  250,  251,  251,  251,  251,  251,  251,
      252,    0,  252,  253,  253,  254,    0,  254,  254,  254,
      255,  255,    0,  255,  255,  255,  256,    0,  256,  257,

      257,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248,  248,  248,  248,  248,  248,
      248,  248,  248,  248,  248
    } ;

/* The intent behind this definition is that it'll catch
 * any uses of REJECT which flex missed.
 */
#define REJECT reject_used_but_not_detected
#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
#line 1 "sln.ll"
#line 6 "sln.ll"

// $Id$
//
//  Copyright (c) 2008, Novartis Institutes for BioMedical Research Inc.
//  All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met: 
//
//     * Redistributions of source code must retain the above copyright 
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above
//       copyright notice, this list of conditions and the following 
//       disclaimer in the documentation and/or other materials provided 
//       with the distribution.
//     * Neither the name of Novartis Institutes for BioMedical Research Inc. 
//       nor the names of its contributors may be used to endorse or promote 
//       products derived from this software without specific prior
//       written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Created by Greg Landrum, September 2006
//

#if defined(__CYGWIN__) && !defined(fileno)
// -std=c++11 turns off recent posix features
extern "C" int fileno(FILE*);
#endif

#include <cstdio>
#ifdef WIN32
#include <io.h> 	 
#endif

#include <RDGeneral/Exceptions.h>
#include <GraphMol/RDKitBase.h>
#include <GraphMol/RDKitQueries.h>
#include <GraphMol/SLNParse/SLNAttribs.h>

#include <string>
#include <cstring>
#include "sln.tab.hpp"

using namespace RDKit;

void setup_sln_string(const std::string &text,yyscan_t yyscanner){
  YY_BUFFER_STATE buff=yysln__scan_string(text.c_str(),yyscanner);
  POSTCONDITION(buff,"invalid buffer");
}

#define YY_FATAL_ERROR(msg) sln_lexer_error(msg)
void sln_lexer_error(const char *msg) {
     BOOST_LOG(rdErrorLog) << msg<<std::endl;
     throw ValueErrorException(msg);
}







#line 758 "/Users/landrgr1/RDKit_git/Code/GraphMol/SLNParse/lex.yysln.cpp"

#define INITIAL 0
#define IN_SLN_PARAM_STATE 1
#define IN_CTAB_PARAM_NAME_STATE 2
#define IN_CTAB_PARAM_VAL_STATE 3
#define IN_PROP_VAL_STATE 4
#define IN_RECURSE_STATE 5

#ifndef YY_NO_UNISTD_H
/* Special case for "unistd.h", since it is non-ANSI. We include it way
 * down here because we want the user's section 1 to have been scanned first.
 * The user has a chance to override it with an option.
 */
#include <unistd.h>
#endif

#ifndef YY_EXTRA_TYPE
#define YY_EXTRA_TYPE void *
#endif

/* Holds the entire state of the reentrant scanner. */
struct yyguts_t
    {

    /* User-defined. Not touched by flex. */
    YY_EXTRA_TYPE yyextra_r;

    /* The rest are the same as the globals declared in the non-reentrant scanner. */
    FILE *yyin_r, *yyout_r;
    size_t yy_buffer_stack_top; /**< index of top of stack. */
    size_t yy_buffer_stack_max; /**< capacity of stack. */
    YY_BUFFER_STATE * yy_buffer_stack; /**< Stack as an array. */
    char yy_hold_char;
    yy_size_t yy_n_chars;
    yy_size_t yyleng_r;
    char *yy_c_buf_p;
    int yy_init;
    int yy_start;
    int yy_did_buffer_switch_on_eof;
    int yy_start_stack_ptr;
    int yy_start_stack_depth;
    int *yy_start_stack;
    yy_state_type yy_last_accepting_state;
    char* yy_last_accepting_cpos;

    int yylineno_r;
    int yy_flex_debug_r;

    char *yytext_r;
    int yy_more_flag;
    int yy_more_len;

    YYSTYPE * yylval_r;

    }; /* end struct yyguts_t */

static int yy_init_globals (yyscan_t yyscanner );

    /* This must go here because YYSTYPE and YYLTYPE are included
     * from bison output in section 1.*/
    #    define yylval yyg->yylval_r
    
int yysln_lex_init (yyscan_t* scanner);

int yysln_lex_init_extra (YY_EXTRA_TYPE user_defined,yyscan_t* scanner);

/* Accessor methods to globals.
   These are made visible to non-reentrant scanners for convenience. */

int yysln_lex_destroy (yyscan_t yyscanner );

int yysln_get_debug (yyscan_t yyscanner );

void yysln_set_debug (int debug_flag ,yyscan_t yyscanner );

YY_EXTRA_TYPE yysln_get_extra (yyscan_t yyscanner );

void yysln_set_extra (YY_EXTRA_TYPE user_defined ,yyscan_t yyscanner );

FILE *yysln_get_in (yyscan_t yyscanner );

void yysln_set_in  (FILE * in_str ,yyscan_t yyscanner );

FILE *yysln_get_out (yyscan_t yyscanner );

void yysln_set_out  (FILE * out_str ,yyscan_t yyscanner );

yy_size_t yysln_get_leng (yyscan_t yyscanner );

char *yysln_get_text (yyscan_t yyscanner );

int yysln_get_lineno (yyscan_t yyscanner );

void yysln_set_lineno (int line_number ,yyscan_t yyscanner );

YYSTYPE * yysln_get_lval (yyscan_t yyscanner );

void yysln_set_lval (YYSTYPE * yylval_param ,yyscan_t yyscanner );

/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int yysln_wrap (yyscan_t yyscanner );
#else
extern int yysln_wrap (yyscan_t yyscanner );
#endif
#endif

    static void yyunput (int c,char *buf_ptr  ,yyscan_t yyscanner);
    
#ifndef yytext_ptr
static void yy_flex_strncpy (char *,yyconst char *,int ,yyscan_t yyscanner);
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (yyconst char * ,yyscan_t yyscanner);
#endif

#ifndef YY_NO_INPUT

#ifdef __cplusplus
static int yyinput (yyscan_t yyscanner );
#else
static int input (yyscan_t yyscanner );
#endif

#endif

    static void yy_push_state (int new_state ,yyscan_t yyscanner);
    
    static void yy_pop_state (yyscan_t yyscanner );
    
    static int yy_top_state (yyscan_t yyscanner );
    
/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#define YY_READ_BUF_SIZE 8192
#endif

/* Copy whatever the last rule matched to the standard output. */
#ifndef ECHO
/* This used to be an fputs(), but since the string might contain NUL's,
 * we now use fwrite().
 */
#define ECHO fwrite( yytext, yyleng, 1, yyout )
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if ( YY_CURRENT_BUFFER_LVALUE->yy_is_interactive ) \
		{ \
		int c = '*'; \
		yy_size_t n; \
		for ( n = 0; n < max_size && \
			     (c = getc( yyin )) != EOF && c != '\n'; ++n ) \
			buf[n] = (char) c; \
		if ( c == '\n' ) \
			buf[n++] = (char) c; \
		if ( c == EOF && ferror( yyin ) ) \
			YY_FATAL_ERROR( "input in flex scanner failed" ); \
		result = n; \
		} \
	else \
		{ \
		errno=0; \
		while ( (result = fread(buf, 1, max_size, yyin))==0 && ferror(yyin)) \
			{ \
			if( errno != EINTR) \
				{ \
				YY_FATAL_ERROR( "input in flex scanner failed" ); \
				break; \
				} \
			errno=0; \
			clearerr(yyin); \
			} \
		}\
\

#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) yy_fatal_error( msg , yyscanner)
#endif

/* end tables serialization structures and prototypes */

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL_IS_OURS 1

extern int yysln_lex \
               (YYSTYPE * yylval_param ,yyscan_t yyscanner);

#define YY_DECL int yysln_lex \
               (YYSTYPE * yylval_param , yyscan_t yyscanner)
#endif /* !YY_DECL */

/* Code executed at the beginning of each rule, after yytext and yyleng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK break;
#endif

#define YY_RULE_SETUP \
	YY_USER_ACTION

/** The main scanner function which does all the work.
 */
YY_DECL
{
	yy_state_type yy_current_state;
	char *yy_cp, *yy_bp;
	int yy_act;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

#line 78 "sln.ll"


#line 1005 "/Users/landrgr1/RDKit_git/Code/GraphMol/SLNParse/lex.yysln.cpp"

    yylval = yylval_param;

	if ( !yyg->yy_init )
		{
		yyg->yy_init = 1;

#ifdef YY_USER_INIT
		YY_USER_INIT;
#endif

		if ( ! yyg->yy_start )
			yyg->yy_start = 1;	/* first start state */

		if ( ! yyin )
			yyin = stdin;

		if ( ! yyout )
			yyout = stdout;

		if ( ! YY_CURRENT_BUFFER ) {
			yysln_ensure_buffer_stack (yyscanner);
			YY_CURRENT_BUFFER_LVALUE =
				yysln__create_buffer(yyin,YY_BUF_SIZE ,yyscanner);
		}

		yysln__load_buffer_state(yyscanner );
		}

	while ( 1 )		/* loops until end-of-file is reached */
		{
		yy_cp = yyg->yy_c_buf_p;

		/* Support of yytext. */
		*yy_cp = yyg->yy_hold_char;

		/* yy_bp points to the position in yy_ch_buf of the start of
		 * the current run.
		 */
		yy_bp = yy_cp;

		yy_current_state = yyg->yy_start;
yy_match:
		do
			{
			YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)];
			if ( yy_accept[yy_current_state] )
				{
				yyg->yy_last_accepting_state = yy_current_state;
				yyg->yy_last_accepting_cpos = yy_cp;
				}
			while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
				{
				yy_current_state = (int) yy_def[yy_current_state];
				if ( yy_current_state >= 249 )
					yy_c = yy_meta[(unsigned int) yy_c];
				}
			yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
			++yy_cp;
			}
		while ( yy_base[yy_current_state] != 702 );

yy_find_action:
		yy_act = yy_accept[yy_current_state];
		if ( yy_act == 0 )
			{ /* have to back up */
			yy_cp = yyg->yy_last_accepting_cpos;
			yy_current_state = yyg->yy_last_accepting_state;
			yy_act = yy_accept[yy_current_state];
			}

		YY_DO_BEFORE_ACTION;

do_action:	/* This label is used only to access EOF actions. */

		switch ( yy_act )
	{ /* beginning of action switch */
			case 0: /* must back up */
			/* undo the effects of YY_DO_BEFORE_ACTION */
			*yy_cp = yyg->yy_hold_char;
			yy_cp = yyg->yy_last_accepting_cpos;
			yy_current_state = yyg->yy_last_accepting_state;
			goto yy_find_action;

case 1:
/* rule 1 can match eol */
YY_RULE_SETUP
#line 80 "sln.ll"
{
  yylval->text_T=new std::string(yytext);
  return TEXT_BLOCK; 
}
	YY_BREAK
case 2:
YY_RULE_SETUP
#line 85 "sln.ll"
{ 
  yylval->text_T=new std::string(yytext);
  return TEXT_BLOCK; 
}
	YY_BREAK
case 3:
YY_RULE_SETUP
#line 90 "sln.ll"
{ 
  yylval->text_T=new std::string(yytext);
  return TEXT_BLOCK; 
}
	YY_BREAK
case 4:
YY_RULE_SETUP
#line 95 "sln.ll"
{ 
  yylval->text_T=new std::string(yytext);
  return TEXT_BLOCK; 
}
	YY_BREAK
case 5:
#line 101 "sln.ll"
case 6:
#line 102 "sln.ll"
case 7:
#line 103 "sln.ll"
case 8:
#line 104 "sln.ll"
case 9:
#line 105 "sln.ll"
case 10:
#line 106 "sln.ll"
case 11:
#line 107 "sln.ll"
case 12:
#line 108 "sln.ll"
case 13:
#line 109 "sln.ll"
case 14:
#line 110 "sln.ll"
case 15:
#line 111 "sln.ll"
case 16:
#line 112 "sln.ll"
case 17:
#line 113 "sln.ll"
case 18:
#line 114 "sln.ll"
case 19:
#line 115 "sln.ll"
case 20:
#line 116 "sln.ll"
case 21:
#line 117 "sln.ll"
case 22:
#line 118 "sln.ll"
case 23:
#line 119 "sln.ll"
case 24:
#line 120 "sln.ll"
case 25:
#line 121 "sln.ll"
case 26:
#line 122 "sln.ll"
case 27:
#line 123 "sln.ll"
case 28:
#line 124 "sln.ll"
case 29:
#line 125 "sln.ll"
case 30:
#line 126 "sln.ll"
case 31:
#line 127 "sln.ll"
case 32:
#line 128 "sln.ll"
case 33:
#line 129 "sln.ll"
case 34:
#line 130 "sln.ll"
case 35:
#line 131 "sln.ll"
case 36:
#line 132 "sln.ll"
case 37:
#line 133 "sln.ll"
case 38:
#line 134 "sln.ll"
case 39:
#line 135 "sln.ll"
case 40:
#line 136 "sln.ll"
case 41:
#line 137 "sln.ll"
case 42:
#line 138 "sln.ll"
case 43:
#line 139 "sln.ll"
case 44:
#line 140 "sln.ll"
case 45:
#line 141 "sln.ll"
case 46:
#line 142 "sln.ll"
case 47:
#line 143 "sln.ll"
case 48:
#line 144 "sln.ll"
case 49:
#line 145 "sln.ll"
case 50:
#line 146 "sln.ll"
case 51:
#line 147 "sln.ll"
case 52:
#line 148 "sln.ll"
case 53:
#line 149 "sln.ll"
case 54:
#line 150 "sln.ll"
case 55:
#line 151 "sln.ll"
case 56:
#line 152 "sln.ll"
case 57:
#line 153 "sln.ll"
case 58:
#line 154 "sln.ll"
case 59:
#line 155 "sln.ll"
case 60:
#line 156 "sln.ll"
case 61:
#line 157 "sln.ll"
case 62:
#line 158 "sln.ll"
case 63:
#line 159 "sln.ll"
case 64:
#line 160 "sln.ll"
case 65:
#line 161 "sln.ll"
case 66:
#line 162 "sln.ll"
case 67:
#line 163 "sln.ll"
case 68:
#line 164 "sln.ll"
case 69:
#line 165 "sln.ll"
case 70:
#line 166 "sln.ll"
case 71:
#line 167 "sln.ll"
case 72:
#line 168 "sln.ll"
case 73:
#line 169 "sln.ll"
case 74:
#line 170 "sln.ll"
case 75:
#line 171 "sln.ll"
case 76:
#line 172 "sln.ll"
case 77:
#line 173 "sln.ll"
case 78:
#line 174 "sln.ll"
case 79:
#line 175 "sln.ll"
case 80:
#line 176 "sln.ll"
case 81:
#line 177 "sln.ll"
case 82:
#line 178 "sln.ll"
case 83:
#line 179 "sln.ll"
case 84:
#line 180 "sln.ll"
case 85:
#line 181 "sln.ll"
case 86:
#line 182 "sln.ll"
case 87:
#line 183 "sln.ll"
case 88:
#line 184 "sln.ll"
case 89:
#line 185 "sln.ll"
case 90:
#line 186 "sln.ll"
case 91:
#line 187 "sln.ll"
case 92:
#line 188 "sln.ll"
case 93:
#line 189 "sln.ll"
case 94:
#line 190 "sln.ll"
case 95:
#line 191 "sln.ll"
case 96:
#line 192 "sln.ll"
case 97:
#line 193 "sln.ll"
case 98:
#line 194 "sln.ll"
case 99:
#line 195 "sln.ll"
case 100:
#line 196 "sln.ll"
case 101:
#line 197 "sln.ll"
case 102:
#line 198 "sln.ll"
case 103:
#line 199 "sln.ll"
case 104:
#line 200 "sln.ll"
case 105:
#line 201 "sln.ll"
case 106:
YY_RULE_SETUP
#line 201 "sln.ll"
{
  if((bool)yyextra){
          yylval->atom_T = new QueryAtom(PeriodicTable::getTable()->getAtomicNumber(yytext));
        } else {
          yylval->atom_T = new Atom(PeriodicTable::getTable()->getAtomicNumber(yytext));
        }
  // SLN has no concept of implicit Hs... they're either in the SLN or they don't exist:        
  yylval->atom_T->setNoImplicit(true);
        
  return ATOM_TOKEN;
}
	YY_BREAK
case 107:
YY_RULE_SETUP
#line 212 "sln.ll"
{
  yylval->atom_T = new QueryAtom();
  yylval->atom_T->setQuery(makeAtomNullQuery());
  // SLN has no concept of implicit Hs... they're either in the SLN or they don't exist:        
  yylval->atom_T->setNoImplicit(true);
  return ATOM_TOKEN;
}
	YY_BREAK
case 108:
YY_RULE_SETUP
#line 219 "sln.ll"
{
  yylval->atom_T = new QueryAtom();
  yylval->atom_T->setQuery(makeAtomNumQuery(1));
  // FIX: are 2H or 3H heavy atoms or Hs?
  yylval->atom_T->getQuery()->setNegation(true);

  // SLN has no concept of implicit Hs... they're either in the SLN or they don't exist:        
  yylval->atom_T->setNoImplicit(true);
  return ATOM_TOKEN;
}
	YY_BREAK
case 109:
YY_RULE_SETUP
#line 229 "sln.ll"
{
  yylval->atom_T = new QueryAtom();
  yylval->atom_T->setQuery(makeAtomNumQuery(9));
  yylval->atom_T->expandQuery(makeAtomNumQuery(17),Queries::COMPOSITE_OR,true);
  yylval->atom_T->expandQuery(makeAtomNumQuery(35),Queries::COMPOSITE_OR,true);
  yylval->atom_T->expandQuery(makeAtomNumQuery(53),Queries::COMPOSITE_OR,true);

  // SLN has no concept of implicit Hs... they're either in the SLN or they don't exist:        
  yylval->atom_T->setNoImplicit(true);
  return ATOM_TOKEN;
}
	YY_BREAK
case 110:
YY_RULE_SETUP
#line 240 "sln.ll"
{
  yylval->atom_T = new QueryAtom();
  yylval->atom_T->setQuery(makeAtomNumQuery(6));
  yylval->atom_T->expandQuery(makeAtomNumQuery(1),Queries::COMPOSITE_OR,true);
  yylval->atom_T->getQuery()->setNegation(true);

  // SLN has no concept of implicit Hs... they're either in the SLN or they don't exist:        
  yylval->atom_T->setNoImplicit(true);
  return ATOM_TOKEN;
}
	YY_BREAK
case 111:
YY_RULE_SETUP
#line 251 "sln.ll"
{ yy_push_state(IN_SLN_PARAM_STATE,yyscanner); return H_BRACKET_TOKEN; }
	YY_BREAK
case 112:
YY_RULE_SETUP
#line 252 "sln.ll"
{ return H_ASTERIX_TOKEN; }
	YY_BREAK
case 113:
YY_RULE_SETUP
#line 253 "sln.ll"
{ return H_TOKEN; }
	YY_BREAK
case 114:
#line 256 "sln.ll"
case 115:
#line 257 "sln.ll"
case 116:
#line 258 "sln.ll"
case 117:
YY_RULE_SETUP
#line 258 "sln.ll"
{ yy_push_state(IN_RECURSE_STATE,yyscanner); return RECURSE_TOKEN; }
	YY_BREAK
case 118:
#line 260 "sln.ll"
case 119:
#line 261 "sln.ll"
case 120:
#line 262 "sln.ll"
case 121:
#line 263 "sln.ll"
case 122:
#line 264 "sln.ll"
case 123:
#line 265 "sln.ll"
case 124:
#line 266 "sln.ll"
case 125:
YY_RULE_SETUP
#line 266 "sln.ll"
{ yy_push_state(IN_RECURSE_STATE,yyscanner); return NEG_RECURSE_TOKEN; }
	YY_BREAK
case 126:
YY_RULE_SETUP
#line 269 "sln.ll"
{ return MINUS_TOKEN; }
	YY_BREAK
case 127:
YY_RULE_SETUP
#line 271 "sln.ll"
{ return PLUS_TOKEN; }
	YY_BREAK
case 128:
YY_RULE_SETUP
#line 272 "sln.ll"
{ return HASH_TOKEN; }  
	YY_BREAK
case 129:
YY_RULE_SETUP
#line 273 "sln.ll"
{ return TILDE_TOKEN; } 
	YY_BREAK
case 130:
YY_RULE_SETUP
#line 274 "sln.ll"
{ return COLON_TOKEN; } 
	YY_BREAK
case 131:
YY_RULE_SETUP
#line 276 "sln.ll"
{ return OPEN_PAREN_TOKEN; }
	YY_BREAK
case 132:
YY_RULE_SETUP
#line 277 "sln.ll"
{ return CLOSE_PAREN_TOKEN; }
	YY_BREAK
case 133:
#line 280 "sln.ll"
case 134:
#line 281 "sln.ll"
case 135:
#line 282 "sln.ll"
case 136:
#line 283 "sln.ll"
case 137:
#line 284 "sln.ll"
case 138:
YY_RULE_SETUP
#line 284 "sln.ll"
{
  yy_push_state(IN_PROP_VAL_STATE,yyscanner);
  yylval->text_T=new std::string(yytext);
  return COMPARE_TOKEN; 
}
	YY_BREAK
case 139:
YY_RULE_SETUP
#line 290 "sln.ll"
{
  yy_pop_state(yyscanner);
  yy_push_state(IN_CTAB_PARAM_VAL_STATE,yyscanner);
  return EQUALS_TOKEN; 
}
	YY_BREAK
case 140:
YY_RULE_SETUP
#line 295 "sln.ll"
{
  yy_pop_state(yyscanner);
  yy_push_state(IN_CTAB_PARAM_VAL_STATE,yyscanner);
  return COLON_EQUALS_TOKEN; 
}
	YY_BREAK
case 141:
YY_RULE_SETUP
#line 300 "sln.ll"
{
  yy_pop_state(yyscanner);
  yy_push_state(IN_CTAB_PARAM_VAL_STATE,yyscanner);
  return CARET_EQUALS_TOKEN; 
}
	YY_BREAK
case 142:
YY_RULE_SETUP
#line 306 "sln.ll"
{
 return EQUALS_TOKEN; 
}
	YY_BREAK
case 143:
YY_RULE_SETUP
#line 312 "sln.ll"
{
  yy_pop_state(yyscanner); 
  return SEMI_TOKEN; 
}
	YY_BREAK
case 144:
YY_RULE_SETUP
#line 316 "sln.ll"
{
  yy_pop_state(yyscanner); 
  return AND_TOKEN; 
}
	YY_BREAK
case 145:
YY_RULE_SETUP
#line 320 "sln.ll"
{  
  return SEMI_TOKEN;
 }
	YY_BREAK
case 146:
YY_RULE_SETUP
#line 323 "sln.ll"
{
  return AND_TOKEN;
 }
	YY_BREAK
case 147:
YY_RULE_SETUP
#line 326 "sln.ll"
{
 yy_pop_state(yyscanner);
 yy_push_state(IN_CTAB_PARAM_NAME_STATE,yyscanner);
 return SEMI_TOKEN; 
}
	YY_BREAK
case 148:
YY_RULE_SETUP
#line 331 "sln.ll"
{
 yy_pop_state(yyscanner);
 yy_push_state(IN_CTAB_PARAM_NAME_STATE,yyscanner);
 return AND_TOKEN; 
}
	YY_BREAK
case 149:
YY_RULE_SETUP
#line 337 "sln.ll"
{
  yy_pop_state(yyscanner);
  return SEMI_TOKEN;
 }
	YY_BREAK
case 150:
YY_RULE_SETUP
#line 341 "sln.ll"
{
  yy_pop_state(yyscanner);
  return AND_TOKEN;
 }
	YY_BREAK
case 151:
YY_RULE_SETUP
#line 345 "sln.ll"
{
  yy_pop_state(yyscanner);
  return OR_TOKEN;
 }
	YY_BREAK
case 152:
YY_RULE_SETUP
#line 349 "sln.ll"
{
  return SEMI_TOKEN;
}
	YY_BREAK
case 153:
YY_RULE_SETUP
#line 352 "sln.ll"
{
  return AND_TOKEN;
}
	YY_BREAK
case 154:
YY_RULE_SETUP
#line 355 "sln.ll"
{
  return NOT_TOKEN;
}
	YY_BREAK
case 155:
YY_RULE_SETUP
#line 360 "sln.ll"
{ yy_push_state(IN_SLN_PARAM_STATE,yyscanner); return OPEN_BRACKET_TOKEN; }
	YY_BREAK
case 156:
YY_RULE_SETUP
#line 362 "sln.ll"
{
	// we're closing a recursive definition, which means we should also be
	//  closing a parameter block:
  yy_pop_state(yyscanner);
	if(YY_START!=IN_SLN_PARAM_STATE){
		std::cerr << " after closing a recursion, we were not in the appropriate state."  <<std::endl;
	} else {
    yy_pop_state(yyscanner);
  }
	return CLOSE_BRACKET_TOKEN;
}
	YY_BREAK
case 157:
YY_RULE_SETUP
#line 373 "sln.ll"
{
  // if we're currently in an SLN property block (e.g. in []'s), we need
  // to pop both the prop_val state and the property block state:
  yy_pop_state(yyscanner);
  if(YY_START==IN_SLN_PARAM_STATE) {
    yy_pop_state(yyscanner);
  } 
  return CLOSE_BRACKET_TOKEN;
}
	YY_BREAK
case 158:
YY_RULE_SETUP
#line 383 "sln.ll"
{ yy_pop_state(yyscanner); return CLOSE_BRACKET_TOKEN; }
	YY_BREAK
case 159:
YY_RULE_SETUP
#line 385 "sln.ll"
{ yy_push_state(IN_CTAB_PARAM_NAME_STATE,yyscanner); return OPEN_ANGLE_TOKEN; }
	YY_BREAK
case 160:
YY_RULE_SETUP
#line 386 "sln.ll"
{ yy_pop_state(yyscanner); return CLOSE_ANGLE_TOKEN; }
	YY_BREAK
case 161:
YY_RULE_SETUP
#line 387 "sln.ll"
{ yy_pop_state(yyscanner); return CLOSE_ANGLE_TOKEN; }
	YY_BREAK
case 162:
YY_RULE_SETUP
#line 388 "sln.ll"
{ 
  // if we're currently in a CTAB property block (e.g. in <>'s), we need
  // to pop both the prop_val state and the property block state:
  yy_pop_state(yyscanner);
  if(YY_START==IN_CTAB_PARAM_VAL_STATE) {
    yy_pop_state(yyscanner);
  } 
  return CLOSE_ANGLE_TOKEN; 
}
	YY_BREAK
case 163:
YY_RULE_SETUP
#line 399 "sln.ll"
{ return SEPARATOR_TOKEN; }
	YY_BREAK
case 164:
YY_RULE_SETUP
#line 400 "sln.ll"
{ 
  return COMMA_TOKEN; 
}
	YY_BREAK
case 165:
YY_RULE_SETUP
#line 404 "sln.ll"
{ return AT_TOKEN; }
	YY_BREAK
case 166:
YY_RULE_SETUP
#line 405 "sln.ll"
{ return ASTERIX_TOKEN; }
	YY_BREAK
case 167:
YY_RULE_SETUP
#line 407 "sln.ll"
{ yylval->ival_T = atoi( yytext ); return DIGIT_TOKEN; }
	YY_BREAK
case 168:
/* rule 168 can match eol */
YY_RULE_SETUP
#line 410 "sln.ll"
return 0;
	YY_BREAK
case YY_STATE_EOF(INITIAL):
case YY_STATE_EOF(IN_SLN_PARAM_STATE):
case YY_STATE_EOF(IN_CTAB_PARAM_NAME_STATE):
case YY_STATE_EOF(IN_CTAB_PARAM_VAL_STATE):
case YY_STATE_EOF(IN_PROP_VAL_STATE):
case YY_STATE_EOF(IN_RECURSE_STATE):
#line 411 "sln.ll"
{ return EOS_TOKEN; }
	YY_BREAK
case 169:
YY_RULE_SETUP
#line 412 "sln.ll"
return yytext[0];
	YY_BREAK
case 170:
YY_RULE_SETUP
#line 414 "sln.ll"
ECHO;
	YY_BREAK
#line 1738 "/Users/landrgr1/RDKit_git/Code/GraphMol/SLNParse/lex.yysln.cpp"

	case YY_END_OF_BUFFER:
		{
		/* Amount of text matched not including the EOB char. */
		int yy_amount_of_matched_text = (int) (yy_cp - yyg->yytext_ptr) - 1;

		/* Undo the effects of YY_DO_BEFORE_ACTION. */
		*yy_cp = yyg->yy_hold_char;
		YY_RESTORE_YY_MORE_OFFSET

		if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_NEW )
			{
			/* We're scanning a new file or input source.  It's
			 * possible that this happened because the user
			 * just pointed yyin at a new source and called
			 * yysln_lex().  If so, then we have to assure
			 * consistency between YY_CURRENT_BUFFER and our
			 * globals.  Here is the right place to do so, because
			 * this is the first action (other than possibly a
			 * back-up) that will match for the new input source.
			 */
			yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
			YY_CURRENT_BUFFER_LVALUE->yy_input_file = yyin;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status = YY_BUFFER_NORMAL;
			}

		/* Note that here we test for yy_c_buf_p "<=" to the position
		 * of the first EOB in the buffer, since yy_c_buf_p will
		 * already have been incremented past the NUL character
		 * (since all states make transitions on EOB to the
		 * end-of-buffer state).  Contrast this with the test
		 * in input().
		 */
		if ( yyg->yy_c_buf_p <= &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] )
			{ /* This was really a NUL. */
			yy_state_type yy_next_state;

			yyg->yy_c_buf_p = yyg->yytext_ptr + yy_amount_of_matched_text;

			yy_current_state = yy_get_previous_state( yyscanner );

			/* Okay, we're now positioned to make the NUL
			 * transition.  We couldn't have
			 * yy_get_previous_state() go ahead and do it
			 * for us because it doesn't know how to deal
			 * with the possibility of jamming (and we don't
			 * want to build jamming into it because then it
			 * will run more slowly).
			 */

			yy_next_state = yy_try_NUL_trans( yy_current_state , yyscanner);

			yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;

			if ( yy_next_state )
				{
				/* Consume the NUL. */
				yy_cp = ++yyg->yy_c_buf_p;
				yy_current_state = yy_next_state;
				goto yy_match;
				}

			else
				{
				yy_cp = yyg->yy_c_buf_p;
				goto yy_find_action;
				}
			}

		else switch ( yy_get_next_buffer( yyscanner ) )
			{
			case EOB_ACT_END_OF_FILE:
				{
				yyg->yy_did_buffer_switch_on_eof = 0;

				if ( yysln_wrap(yyscanner ) )
					{
					/* Note: because we've taken care in
					 * yy_get_next_buffer() to have set up
					 * yytext, we can now set up
					 * yy_c_buf_p so that if some total
					 * hoser (like flex itself) wants to
					 * call the scanner after we return the
					 * YY_NULL, it'll still work - another
					 * YY_NULL will get returned.
					 */
					yyg->yy_c_buf_p = yyg->yytext_ptr + YY_MORE_ADJ;

					yy_act = YY_STATE_EOF(YY_START);
					goto do_action;
					}

				else
					{
					if ( ! yyg->yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
					}
				break;
				}

			case EOB_ACT_CONTINUE_SCAN:
				yyg->yy_c_buf_p =
					yyg->yytext_ptr + yy_amount_of_matched_text;

				yy_current_state = yy_get_previous_state( yyscanner );

				yy_cp = yyg->yy_c_buf_p;
				yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
				goto yy_match;

			case EOB_ACT_LAST_MATCH:
				yyg->yy_c_buf_p =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars];

				yy_current_state = yy_get_previous_state( yyscanner );

				yy_cp = yyg->yy_c_buf_p;
				yy_bp = yyg->yytext_ptr + YY_MORE_ADJ;
				goto yy_find_action;
			}
		break;
		}

	default:
		YY_FATAL_ERROR(
			"fatal flex scanner internal error--no action found" );
	} /* end of action switch */
		} /* end of scanning one token */
} /* end of yysln_lex */

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */
static int yy_get_next_buffer (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	char *dest = YY_CURRENT_BUFFER_LVALUE->yy_ch_buf;
	char *source = yyg->yytext_ptr;
	int number_to_move, i;
	int ret_val;

	if ( yyg->yy_c_buf_p > &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] )
		YY_FATAL_ERROR(
		"fatal flex scanner internal error--end of buffer missed" );

	if ( YY_CURRENT_BUFFER_LVALUE->yy_fill_buffer == 0 )
		{ /* Don't try to fill the buffer, so this is an EOF. */
		if ( yyg->yy_c_buf_p - yyg->yytext_ptr - YY_MORE_ADJ == 1 )
			{
			/* We matched a single character, the EOB, so
			 * treat this as a final EOF.
			 */
			return EOB_ACT_END_OF_FILE;
			}

		else
			{
			/* We matched some text prior to the EOB, first
			 * process it.
			 */
			return EOB_ACT_LAST_MATCH;
			}
		}

	/* Try to read more data. */

	/* First move last chars to start of buffer. */
	number_to_move = (int) (yyg->yy_c_buf_p - yyg->yytext_ptr) - 1;

	for ( i = 0; i < number_to_move; ++i )
		*(dest++) = *(source++);

	if ( YY_CURRENT_BUFFER_LVALUE->yy_buffer_status == YY_BUFFER_EOF_PENDING )
		/* don't do the read, it's not guaranteed to return an EOF,
		 * just force an EOF
		 */
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars = 0;

	else
		{
			yy_size_t num_to_read =
			YY_CURRENT_BUFFER_LVALUE->yy_buf_size - number_to_move - 1;

		while ( num_to_read <= 0 )
			{ /* Not enough room in the buffer - grow it. */

			/* just a shorter name for the current buffer */
			YY_BUFFER_STATE b = YY_CURRENT_BUFFER;

			int yy_c_buf_p_offset =
				(int) (yyg->yy_c_buf_p - b->yy_ch_buf);

			if ( b->yy_is_our_buffer )
				{
				yy_size_t new_size = b->yy_buf_size * 2;

				if ( new_size <= 0 )
					b->yy_buf_size += b->yy_buf_size / 8;
				else
					b->yy_buf_size *= 2;

				b->yy_ch_buf = (char *)
					/* Include room in for 2 EOB chars. */
					yysln_realloc((void *) b->yy_ch_buf,b->yy_buf_size + 2 ,yyscanner );
				}
			else
				/* Can't grow it, we don't own it. */
				b->yy_ch_buf = nullptr;

			if ( ! b->yy_ch_buf )
				YY_FATAL_ERROR(
				"fatal error - scanner input buffer overflow" );

			yyg->yy_c_buf_p = &b->yy_ch_buf[yy_c_buf_p_offset];

			num_to_read = YY_CURRENT_BUFFER_LVALUE->yy_buf_size -
						number_to_move - 1;

			}

		if ( num_to_read > YY_READ_BUF_SIZE )
			num_to_read = YY_READ_BUF_SIZE;

		/* Read in more data. */
		YY_INPUT( (&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move]),
			yyg->yy_n_chars, num_to_read );

		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	if ( yyg->yy_n_chars == 0 )
		{
		if ( number_to_move == YY_MORE_ADJ )
			{
			ret_val = EOB_ACT_END_OF_FILE;
			yysln_restart(yyin  ,yyscanner);
			}

		else
			{
			ret_val = EOB_ACT_LAST_MATCH;
			YY_CURRENT_BUFFER_LVALUE->yy_buffer_status =
				YY_BUFFER_EOF_PENDING;
			}
		}

	else
		ret_val = EOB_ACT_CONTINUE_SCAN;

	if ((yy_size_t) (yyg->yy_n_chars + number_to_move) > YY_CURRENT_BUFFER_LVALUE->yy_buf_size) {
		/* Extend the array by 50%, plus the number we really need. */
		yy_size_t new_size = yyg->yy_n_chars + number_to_move + (yyg->yy_n_chars >> 1);
		YY_CURRENT_BUFFER_LVALUE->yy_ch_buf = (char *) yysln_realloc((void *) YY_CURRENT_BUFFER_LVALUE->yy_ch_buf,new_size ,yyscanner );
		if ( ! YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			YY_FATAL_ERROR( "out of dynamic memory in yy_get_next_buffer()" );
	}

	yyg->yy_n_chars += number_to_move;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] = YY_END_OF_BUFFER_CHAR;
	YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars + 1] = YY_END_OF_BUFFER_CHAR;

	yyg->yytext_ptr = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[0];

	return ret_val;
}

/* yy_get_previous_state - get the state just before the EOB char was reached */

    static yy_state_type yy_get_previous_state (yyscan_t yyscanner)
{
	yy_state_type yy_current_state;
	char *yy_cp;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	yy_current_state = yyg->yy_start;

	for ( yy_cp = yyg->yytext_ptr + YY_MORE_ADJ; yy_cp < yyg->yy_c_buf_p; ++yy_cp )
		{
		YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 1);
		if ( yy_accept[yy_current_state] )
			{
			yyg->yy_last_accepting_state = yy_current_state;
			yyg->yy_last_accepting_cpos = yy_cp;
			}
		while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
			{
			yy_current_state = (int) yy_def[yy_current_state];
			if ( yy_current_state >= 249 )
				yy_c = yy_meta[(unsigned int) yy_c];
			}
		yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
		}

	return yy_current_state;
}

/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */
    static yy_state_type yy_try_NUL_trans  (yy_state_type yy_current_state , yyscan_t yyscanner)
{
	int yy_is_jam;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner; /* This var may be unused depending upon options. */
	char *yy_cp = yyg->yy_c_buf_p;

	YY_CHAR yy_c = 1;
	if ( yy_accept[yy_current_state] )
		{
		yyg->yy_last_accepting_state = yy_current_state;
		yyg->yy_last_accepting_cpos = yy_cp;
		}
	while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		yy_current_state = (int) yy_def[yy_current_state];
		if ( yy_current_state >= 249 )
			yy_c = yy_meta[(unsigned int) yy_c];
		}
	yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
	yy_is_jam = (yy_current_state == 248);

	return yy_is_jam ? 0 : yy_current_state;
}

    static void yyunput (int c, char * yy_bp , yyscan_t yyscanner)
{
	char *yy_cp;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    yy_cp = yyg->yy_c_buf_p;

	/* undo effects of setting up yytext */
	*yy_cp = yyg->yy_hold_char;

	if ( yy_cp < YY_CURRENT_BUFFER_LVALUE->yy_ch_buf + 2 )
		{ /* need to shift things up to make room */
		/* +2 for EOB chars. */
		yy_size_t number_to_move = yyg->yy_n_chars + 2;
		char *dest = &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[
					YY_CURRENT_BUFFER_LVALUE->yy_buf_size + 2];
		char *source =
				&YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[number_to_move];

		while ( source > YY_CURRENT_BUFFER_LVALUE->yy_ch_buf )
			*--dest = *--source;

		yy_cp += (int) (dest - source);
		yy_bp += (int) (dest - source);
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars =
			yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_buf_size;

		if ( yy_cp < YY_CURRENT_BUFFER_LVALUE->yy_ch_buf + 2 )
			YY_FATAL_ERROR( "flex scanner push-back overflow" );
		}

	*--yy_cp = (char) c;

	yyg->yytext_ptr = yy_bp;
	yyg->yy_hold_char = *yy_cp;
	yyg->yy_c_buf_p = yy_cp;
}

#ifndef YY_NO_INPUT
#ifdef __cplusplus
    static int yyinput (yyscan_t yyscanner)
#else
    static int input  (yyscan_t yyscanner)
#endif

{
	int c;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	*yyg->yy_c_buf_p = yyg->yy_hold_char;

	if ( *yyg->yy_c_buf_p == YY_END_OF_BUFFER_CHAR )
		{
		/* yy_c_buf_p now points to the character we want to return.
		 * If this occurs *before* the EOB characters, then it's a
		 * valid NUL; if not, then we've hit the end of the buffer.
		 */
		if ( yyg->yy_c_buf_p < &YY_CURRENT_BUFFER_LVALUE->yy_ch_buf[yyg->yy_n_chars] )
			/* This was really a NUL. */
			*yyg->yy_c_buf_p = '\0';

		else
			{ /* need more input */
			yy_size_t offset = yyg->yy_c_buf_p - yyg->yytext_ptr;
			++yyg->yy_c_buf_p;

			switch ( yy_get_next_buffer( yyscanner ) )
				{
				case EOB_ACT_LAST_MATCH:
					/* This happens because yy_g_n_b()
					 * sees that we've accumulated a
					 * token and flags that we need to
					 * try matching the token before
					 * proceeding.  But for input(),
					 * there's no matching to consider.
					 * So convert the EOB_ACT_LAST_MATCH
					 * to EOB_ACT_END_OF_FILE.
					 */

					/* Reset buffer status. */
					yysln_restart(yyin ,yyscanner);

					/*FALLTHROUGH*/

				case EOB_ACT_END_OF_FILE:
					{
					if ( yysln_wrap(yyscanner ) )
						return 0;

					if ( ! yyg->yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
#ifdef __cplusplus
					return yyinput(yyscanner);
#else
					return input(yyscanner);
#endif
					}

				case EOB_ACT_CONTINUE_SCAN:
					yyg->yy_c_buf_p = yyg->yytext_ptr + offset;
					break;
				}
			}
		}

	c = *(unsigned char *) yyg->yy_c_buf_p;	/* cast for 8-bit char's */
	*yyg->yy_c_buf_p = '\0';	/* preserve yytext */
	yyg->yy_hold_char = *++yyg->yy_c_buf_p;

	return c;
}
#endif	/* ifndef YY_NO_INPUT */

/** Immediately switch to a different input stream.
 * @param input_file A readable stream.
 * @param yyscanner The scanner object.
 * @note This function does not reset the start condition to @c INITIAL .
 */
    void yysln_restart  (FILE * input_file , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if ( ! YY_CURRENT_BUFFER ){
        yysln_ensure_buffer_stack (yyscanner);
		YY_CURRENT_BUFFER_LVALUE =
            yysln__create_buffer(yyin,YY_BUF_SIZE ,yyscanner);
	}

	yysln__init_buffer(YY_CURRENT_BUFFER,input_file ,yyscanner);
	yysln__load_buffer_state(yyscanner );
}

/** Switch to a different input buffer.
 * @param new_buffer The new input buffer.
 * @param yyscanner The scanner object.
 */
    void yysln__switch_to_buffer  (YY_BUFFER_STATE  new_buffer , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	/* TODO. We should be able to replace this entire function body
	 * with
	 *		yysln_pop_buffer_state();
	 *		yysln_push_buffer_state(new_buffer);
     */
	yysln_ensure_buffer_stack (yyscanner);
	if ( YY_CURRENT_BUFFER == new_buffer )
		return;

	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*yyg->yy_c_buf_p = yyg->yy_hold_char;
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	YY_CURRENT_BUFFER_LVALUE = new_buffer;
	yysln__load_buffer_state(yyscanner );

	/* We don't actually know whether we did this switch during
	 * EOF (yysln_wrap()) processing, but the only time this flag
	 * is looked at is after yysln_wrap() is called, so it's safe
	 * to go ahead and always set it.
	 */
	yyg->yy_did_buffer_switch_on_eof = 1;
}

static void yysln__load_buffer_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	yyg->yy_n_chars = YY_CURRENT_BUFFER_LVALUE->yy_n_chars;
	yyg->yytext_ptr = yyg->yy_c_buf_p = YY_CURRENT_BUFFER_LVALUE->yy_buf_pos;
	yyin = YY_CURRENT_BUFFER_LVALUE->yy_input_file;
	yyg->yy_hold_char = *yyg->yy_c_buf_p;
}

/** Allocate and initialize an input buffer state.
 * @param file A readable stream.
 * @param size The character buffer size in bytes. When in doubt, use @c YY_BUF_SIZE.
 * @param yyscanner The scanner object.
 * @return the allocated buffer state.
 */
    YY_BUFFER_STATE yysln__create_buffer  (FILE * file, int  size , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
    
	b = (YY_BUFFER_STATE) yysln_alloc(sizeof( struct yy_buffer_state ) ,yyscanner );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yysln__create_buffer()" );

	b->yy_buf_size = size;

	/* yy_ch_buf has to be 2 characters longer than the size given because
	 * we need to put in 2 end-of-buffer characters.
	 */
	b->yy_ch_buf = (char *) yysln_alloc(b->yy_buf_size + 2 ,yyscanner );
	if ( ! b->yy_ch_buf )
		YY_FATAL_ERROR( "out of dynamic memory in yysln__create_buffer()" );

	b->yy_is_our_buffer = 1;

	yysln__init_buffer(b,file ,yyscanner);

	return b;
}

/** Destroy the buffer.
 * @param b a buffer created with yysln__create_buffer()
 * @param yyscanner The scanner object.
 */
    void yysln__delete_buffer (YY_BUFFER_STATE  b , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if ( ! b )
		return;

	if ( b == YY_CURRENT_BUFFER ) /* Not sure if we should pop here. */
		YY_CURRENT_BUFFER_LVALUE = (YY_BUFFER_STATE) nullptr;

	if ( b->yy_is_our_buffer )
		yysln_free((void *) b->yy_ch_buf ,yyscanner );

	yysln_free((void *) b ,yyscanner );
}

#ifndef __cplusplus
extern int isatty (int );
#endif /* __cplusplus */
    
/* Initializes or reinitializes a buffer.
 * This function is sometimes called more than once on the same buffer,
 * such as during a yysln_restart() or at EOF.
 */
    static void yysln__init_buffer  (YY_BUFFER_STATE  b, FILE * file , yyscan_t yyscanner)

{
	int oerrno = errno;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	yysln__flush_buffer(b ,yyscanner);

	b->yy_input_file = file;
	b->yy_fill_buffer = 1;

    /* If b is the current buffer, then yysln__init_buffer was _probably_
     * called from yysln_restart() or through yy_get_next_buffer.
     * In that case, we don't want to reset the lineno or column.
     */
    if (b != YY_CURRENT_BUFFER){
        b->yy_bs_lineno = 1;
        b->yy_bs_column = 0;
    }

        b->yy_is_interactive = file ? (isatty( fileno(file) ) > 0) : 0;
    
	errno = oerrno;
}

/** Discard all buffered characters. On the next scan, YY_INPUT will be called.
 * @param b the buffer state to be flushed, usually @c YY_CURRENT_BUFFER.
 * @param yyscanner The scanner object.
 */
    void yysln__flush_buffer (YY_BUFFER_STATE  b , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( ! b )
		return;

	b->yy_n_chars = 0;

	/* We always need two end-of-buffer characters.  The first causes
	 * a transition to the end-of-buffer state.  The second causes
	 * a jam in that state.
	 */
	b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
	b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

	b->yy_buf_pos = &b->yy_ch_buf[0];

	b->yy_at_bol = 1;
	b->yy_buffer_status = YY_BUFFER_NEW;

	if ( b == YY_CURRENT_BUFFER )
		yysln__load_buffer_state(yyscanner );
}

/** Pushes the new state onto the stack. The new state becomes
 *  the current state. This function will allocate the stack
 *  if necessary.
 *  @param new_buffer The new state.
 *  @param yyscanner The scanner object.
 */
void yysln_push_buffer_state (YY_BUFFER_STATE new_buffer , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if (new_buffer == nullptr)
		return;

	yysln_ensure_buffer_stack(yyscanner);

	/* This block is copied from yysln__switch_to_buffer. */
	if ( YY_CURRENT_BUFFER )
		{
		/* Flush out information for old buffer. */
		*yyg->yy_c_buf_p = yyg->yy_hold_char;
		YY_CURRENT_BUFFER_LVALUE->yy_buf_pos = yyg->yy_c_buf_p;
		YY_CURRENT_BUFFER_LVALUE->yy_n_chars = yyg->yy_n_chars;
		}

	/* Only push if top exists. Otherwise, replace top. */
	if (YY_CURRENT_BUFFER)
		yyg->yy_buffer_stack_top++;
	YY_CURRENT_BUFFER_LVALUE = new_buffer;

	/* copied from yysln__switch_to_buffer. */
	yysln__load_buffer_state(yyscanner );
	yyg->yy_did_buffer_switch_on_eof = 1;
}

/** Removes and deletes the top of the stack, if present.
 *  The next element becomes the new top.
 *  @param yyscanner The scanner object.
 */
void yysln_pop_buffer_state (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if (!YY_CURRENT_BUFFER)
		return;

	yysln__delete_buffer(YY_CURRENT_BUFFER ,yyscanner);
	YY_CURRENT_BUFFER_LVALUE = nullptr;
	if (yyg->yy_buffer_stack_top > 0)
		--yyg->yy_buffer_stack_top;

	if (YY_CURRENT_BUFFER) {
		yysln__load_buffer_state(yyscanner );
		yyg->yy_did_buffer_switch_on_eof = 1;
	}
}

/* Allocates the stack if it does not exist.
 *  Guarantees space for at least one push.
 */
static void yysln_ensure_buffer_stack (yyscan_t yyscanner)
{
	yy_size_t num_to_alloc;
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

	if (!yyg->yy_buffer_stack) {

		/* First allocation is just for 2 elements, since we don't know if this
		 * scanner will even need a stack. We use 2 instead of 1 to avoid an
		 * immediate realloc on the next call.
         */
		num_to_alloc = 1;
		yyg->yy_buffer_stack = (struct yy_buffer_state**)yysln_alloc
								(num_to_alloc * sizeof(struct yy_buffer_state*)
								, yyscanner);
		if ( ! yyg->yy_buffer_stack )
			YY_FATAL_ERROR( "out of dynamic memory in yysln_ensure_buffer_stack()" );
								  
		memset(yyg->yy_buffer_stack, 0, num_to_alloc * sizeof(struct yy_buffer_state*));
				
		yyg->yy_buffer_stack_max = num_to_alloc;
		yyg->yy_buffer_stack_top = 0;
		return;
	}

	if (yyg->yy_buffer_stack_top >= (yyg->yy_buffer_stack_max) - 1){

		/* Increase the buffer to prepare for a possible push. */
		int grow_size = 8 /* arbitrary grow size */;

		num_to_alloc = yyg->yy_buffer_stack_max + grow_size;
		yyg->yy_buffer_stack = (struct yy_buffer_state**)yysln_realloc
								(yyg->yy_buffer_stack,
								num_to_alloc * sizeof(struct yy_buffer_state*)
								, yyscanner);
		if ( ! yyg->yy_buffer_stack )
			YY_FATAL_ERROR( "out of dynamic memory in yysln_ensure_buffer_stack()" );

		/* zero only the new slots.*/
		memset(yyg->yy_buffer_stack + yyg->yy_buffer_stack_max, 0, grow_size * sizeof(struct yy_buffer_state*));
		yyg->yy_buffer_stack_max = num_to_alloc;
	}
}

/** Setup the input buffer state to scan directly from a user-specified character buffer.
 * @param base the character buffer
 * @param size the size in bytes of the character buffer
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object. 
 */
YY_BUFFER_STATE yysln__scan_buffer  (char * base, yy_size_t  size , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
    
	if ( size < 2 ||
	     base[size-2] != YY_END_OF_BUFFER_CHAR ||
	     base[size-1] != YY_END_OF_BUFFER_CHAR )
		/* They forgot to leave room for the EOB's. */
		return nullptr;

	b = (YY_BUFFER_STATE) yysln_alloc(sizeof( struct yy_buffer_state ) ,yyscanner );
	if ( ! b )
		YY_FATAL_ERROR( "out of dynamic memory in yysln__scan_buffer()" );

	b->yy_buf_size = size - 2;	/* "- 2" to take care of EOB's */
	b->yy_buf_pos = b->yy_ch_buf = base;
	b->yy_is_our_buffer = 0;
	b->yy_input_file = nullptr;
	b->yy_n_chars = b->yy_buf_size;
	b->yy_is_interactive = 0;
	b->yy_at_bol = 1;
	b->yy_fill_buffer = 0;
	b->yy_buffer_status = YY_BUFFER_NEW;

	yysln__switch_to_buffer(b ,yyscanner );

	return b;
}

/** Setup the input buffer state to scan a string. The next call to yysln_lex() will
 * scan from a @e copy of @a str.
 * @param yystr a NUL-terminated string to scan
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 * @note If you want to scan bytes that may contain NUL values, then use
 *       yysln__scan_bytes() instead.
 */
YY_BUFFER_STATE yysln__scan_string (yyconst char * yystr , yyscan_t yyscanner)
{
    
	return yysln__scan_bytes(yystr,strlen(yystr) ,yyscanner);
}

/** Setup the input buffer state to scan the given bytes. The next call to yysln_lex() will
 * scan from a @e copy of @a bytes.
 * @param bytes the byte buffer to scan
 * @param len the number of bytes in the buffer pointed to by @a bytes.
 * @param yyscanner The scanner object.
 * @return the newly allocated buffer state object.
 */
YY_BUFFER_STATE yysln__scan_bytes  (yyconst char * yybytes, yy_size_t  _yybytes_len , yyscan_t yyscanner)
{
	YY_BUFFER_STATE b;
	char *buf;
	yy_size_t n, i;
    
	/* Get memory for full buffer, including space for trailing EOB's. */
	n = _yybytes_len + 2;
	buf = (char *) yysln_alloc(n ,yyscanner );
	if ( ! buf )
		YY_FATAL_ERROR( "out of dynamic memory in yysln__scan_bytes()" );

	for ( i = 0; i < _yybytes_len; ++i )
		buf[i] = yybytes[i];

	buf[_yybytes_len] = buf[_yybytes_len+1] = YY_END_OF_BUFFER_CHAR;

	b = yysln__scan_buffer(buf,n ,yyscanner);
	if ( ! b )
		YY_FATAL_ERROR( "bad buffer in yysln__scan_bytes()" );

	/* It's okay to grow etc. this buffer, and we should throw it
	 * away when we're done.
	 */
	b->yy_is_our_buffer = 1;

	return b;
}

    static void yy_push_state (int  new_state , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( yyg->yy_start_stack_ptr >= yyg->yy_start_stack_depth )
		{
		yy_size_t new_size;

		yyg->yy_start_stack_depth += YY_START_STACK_INCR;
		new_size = yyg->yy_start_stack_depth * sizeof( int );

		if ( ! yyg->yy_start_stack )
			yyg->yy_start_stack = (int *) yysln_alloc(new_size ,yyscanner );

		else
			yyg->yy_start_stack = (int *) yysln_realloc((void *) yyg->yy_start_stack,new_size ,yyscanner );

		if ( ! yyg->yy_start_stack )
			YY_FATAL_ERROR( "out of memory expanding start-condition stack" );
		}

	yyg->yy_start_stack[yyg->yy_start_stack_ptr++] = YY_START;

	BEGIN(new_state);
}

    static void yy_pop_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	if ( --yyg->yy_start_stack_ptr < 0 )
		YY_FATAL_ERROR( "start-condition stack underflow" );

	BEGIN(yyg->yy_start_stack[yyg->yy_start_stack_ptr]);
}

    static int yy_top_state  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
	return yyg->yy_start_stack[yyg->yy_start_stack_ptr - 1];
}

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif

static void yy_fatal_error (yyconst char* msg , yyscan_t yyscanner)
{
    RDUNUSED_PARAM(yyscanner);
  	(void) fprintf( stderr, "%s\n", msg );
	exit( YY_EXIT_FAILURE );
}

/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
        int yyless_macro_arg = (n); \
        YY_LESS_LINENO(yyless_macro_arg);\
		yytext[yyleng] = yyg->yy_hold_char; \
		yyg->yy_c_buf_p = yytext + yyless_macro_arg; \
		yyg->yy_hold_char = *yyg->yy_c_buf_p; \
		*yyg->yy_c_buf_p = '\0'; \
		yyleng = yyless_macro_arg; \
		} \
	while ( 0 )

/* Accessor  methods (get/set functions) to struct members. */

/** Get the user-defined data for this scanner.
 * @param yyscanner The scanner object.
 */
YY_EXTRA_TYPE yysln_get_extra  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyextra;
}

/** Get the current line number.
 * @param yyscanner The scanner object.
 */
int yysln_get_lineno  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    
        if (! YY_CURRENT_BUFFER)
            return 0;
    
    return yylineno;
}

/** Get the current column number.
 * @param yyscanner The scanner object.
 */
int yysln_get_column  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    
        if (! YY_CURRENT_BUFFER)
            return 0;
    
    return yycolumn;
}

/** Get the input stream.
 * @param yyscanner The scanner object.
 */
FILE *yysln_get_in  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyin;
}

/** Get the output stream.
 * @param yyscanner The scanner object.
 */
FILE *yysln_get_out  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyout;
}

/** Get the length of the current token.
 * @param yyscanner The scanner object.
 */
yy_size_t yysln_get_leng  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yyleng;
}

/** Get the current token.
 * @param yyscanner The scanner object.
 */

char *yysln_get_text  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yytext;
}

/** Set the user-defined data. This data is never touched by the scanner.
 * @param user_defined The data to be associated with this scanner.
 * @param yyscanner The scanner object.
 */
void yysln_set_extra (YY_EXTRA_TYPE  user_defined , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyextra = user_defined ;
}

/** Set the current line number.
 * @param line_number
 * @param yyscanner The scanner object.
 */
void yysln_set_lineno (int  line_number , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        /* lineno is only valid if an input buffer exists. */
        if (! YY_CURRENT_BUFFER )
           yy_fatal_error( "yysln_set_lineno called with no buffer" , yyscanner); 
    
    yylineno = line_number;
}

/** Set the current column.
 * @param line_number
 * @param yyscanner The scanner object.
 */
void yysln_set_column (int  column_no , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

        /* column is only valid if an input buffer exists. */
        if (! YY_CURRENT_BUFFER )
           yy_fatal_error( "yysln_set_column called with no buffer" , yyscanner); 
    
    yycolumn = column_no;
}

/** Set the input stream. This does not discard the current
 * input buffer.
 * @param in_str A readable stream.
 * @param yyscanner The scanner object.
 * @see yysln__switch_to_buffer
 */
void yysln_set_in (FILE *  in_str , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyin = in_str ;
}

void yysln_set_out (FILE *  out_str , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yyout = out_str ;
}

int yysln_get_debug  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yy_flex_debug;
}

void yysln_set_debug (int  bdebug , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yy_flex_debug = bdebug ;
}

/* Accessor methods for yylval and yylloc */

YYSTYPE * yysln_get_lval  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    return yylval;
}

void yysln_set_lval (YYSTYPE *  yylval_param , yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    yylval = yylval_param;
}

/* User-visible API */

/* yysln_lex_init is special because it creates the scanner itself, so it is
 * the ONLY reentrant function that doesn't take the scanner as the last argument.
 * That's why we explicitly handle the declaration, instead of using our macros.
 */

int yysln_lex_init(yyscan_t* ptr_yy_globals)

{
    if (ptr_yy_globals == nullptr){
        errno = EINVAL;
        return 1;
    }

    *ptr_yy_globals = (yyscan_t) yysln_alloc ( sizeof( struct yyguts_t ), nullptr );

    if (*ptr_yy_globals == nullptr){
        errno = ENOMEM;
        return 1;
    }

    /* By setting to 0xAA, we expose bugs in yy_init_globals. Leave at 0x00 for releases. */
    memset(*ptr_yy_globals,0x00,sizeof(struct yyguts_t));

    return yy_init_globals ( *ptr_yy_globals );
}

/* yysln_lex_init_extra has the same functionality as yysln_lex_init, but follows the
 * convention of taking the scanner as the last argument. Note however, that
 * this is a *pointer* to a scanner, as it will be allocated by this call (and
 * is the reason, too, why this function also must handle its own declaration).
 * The user defined value in the first argument will be available to yysln_alloc in
 * the yyextra field.
 */

int yysln_lex_init_extra(YY_EXTRA_TYPE yy_user_defined,yyscan_t* ptr_yy_globals )

{
    struct yyguts_t dummy_yyguts;

    yysln_set_extra (yy_user_defined, &dummy_yyguts);

    if (ptr_yy_globals == nullptr){
        errno = EINVAL;
        return 1;
    }
	
    *ptr_yy_globals = (yyscan_t) yysln_alloc ( sizeof( struct yyguts_t ), &dummy_yyguts );
	
    if (*ptr_yy_globals == nullptr){
        errno = ENOMEM;
        return 1;
    }
    
    /* By setting to 0xAA, we expose bugs in
    yy_init_globals. Leave at 0x00 for releases. */
    memset(*ptr_yy_globals,0x00,sizeof(struct yyguts_t));
    
    yysln_set_extra (yy_user_defined, *ptr_yy_globals);
    
    return yy_init_globals ( *ptr_yy_globals );
}

static int yy_init_globals (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;
    /* Initialization is the same as for the non-reentrant scanner.
     * This function is called from yysln_lex_destroy(), so don't allocate here.
     */

    yyg->yy_buffer_stack = nullptr;
    yyg->yy_buffer_stack_top = 0;
    yyg->yy_buffer_stack_max = 0;
    yyg->yy_c_buf_p = (char *) nullptr;
    yyg->yy_init = 0;
    yyg->yy_start = 0;

    yyg->yy_start_stack_ptr = 0;
    yyg->yy_start_stack_depth = 0;
    yyg->yy_start_stack =  nullptr;

/* Defined in main.c */
#ifdef YY_STDINIT
    yyin = stdin;
    yyout = stdout;
#else
    yyin = (FILE *) nullptr;
    yyout = (FILE *) nullptr;
#endif

    /* For future reference: Set errno on error, since we are called by
     * yysln_lex_init()
     */
    return 0;
}

/* yysln_lex_destroy is for both reentrant and non-reentrant scanners. */
int yysln_lex_destroy  (yyscan_t yyscanner)
{
    struct yyguts_t * yyg = (struct yyguts_t*)yyscanner;

    /* Pop the buffer stack, destroying each element. */
	while(YY_CURRENT_BUFFER){
		yysln__delete_buffer(YY_CURRENT_BUFFER ,yyscanner );
		YY_CURRENT_BUFFER_LVALUE = nullptr;
		yysln_pop_buffer_state(yyscanner);
	}

	/* Destroy the stack itself. */
	yysln_free(yyg->yy_buffer_stack ,yyscanner);
	yyg->yy_buffer_stack = nullptr;

    /* Destroy the start condition stack. */
        yysln_free(yyg->yy_start_stack ,yyscanner );
        yyg->yy_start_stack = nullptr;

    /* Reset the globals. This is important in a non-reentrant scanner so the next time
     * yysln_lex() is called, initialization will occur. */
    yy_init_globals( yyscanner);

    /* Destroy the main struct (reentrant only). */
    yysln_free ( yyscanner , yyscanner );
    yyscanner = nullptr;
    return 0;
}

/*
 * Internal utility routines.
 */

#ifndef yytext_ptr
static void yy_flex_strncpy (char* s1, yyconst char * s2, int n , yyscan_t yyscanner)
{
	int i;
	for ( i = 0; i < n; ++i )
		s1[i] = s2[i];
}
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen (yyconst char * s , yyscan_t yyscanner)
{
	int n;
	for ( n = 0; s[n]; ++n )
		;

	return n;
}
#endif

void *yysln_alloc (yy_size_t  size , yyscan_t yyscanner)
{
    RDUNUSED_PARAM(yyscanner);
	return (void *) malloc( size );
}

void *yysln_realloc  (void * ptr, yy_size_t  size , yyscan_t yyscanner)
{
	/* The cast to (char *) in the following accommodates both
	 * implementations that use char* generic pointers, and those
	 * that use void* generic pointers.  It works with the latter
	 * because both ANSI C and C++ allow castless assignment from
	 * any pointer type to void*, and deal with argument conversions
	 * as though doing an assignment.
	 */
    RDUNUSED_PARAM(yyscanner);
	return (void *) realloc( (char *) ptr, size );
}

void yysln_free (void * ptr , yyscan_t yyscanner)
{
    RDUNUSED_PARAM(yyscanner);
	free( (char *) ptr );	/* see yysln_realloc() for (char *) cast */
}

#define YYTABLES_NAME "yytables"

#line 414 "sln.ll"



#undef yysln_wrap
int yysln_wrap( void ) { return 1; }





