//
//  Copyright (C) 2003-2006 Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
#include <RDGeneral/export.h>
#ifndef _RDMOLOPS_INCL_
#define _RDMOLOPS_INCL_

#define PY_ARRAY_UNIQUE_SYMBOL rdmolops_array_API

void rdkit_rdmolops_ensure_numpy();

#endif
