//
//   Copyright (C) 2003-2006 Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
#include <RDGeneral/export.h>
#ifndef __RD_TRANSFORM_H__
#define __RD_TRANSFORM_H__

namespace RDGeom {

typedef enum { X_Axis, Y_Axis, Z_Axis } AxisType;
}

#endif
