// CommonCS/LibCommon/Hdr/CDXMLParser.h
// Copyright (c) 1999-2004, CambridgeSoft Corp., All Rights Reserved

// BSD 3-Clause License
// 
// Copyright (c) 1986-2025, CambridgeSoft Corp, Revvity Inc and others.
// 
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
// 
// 3. Neither the name of the copyright holder nor the names of its
//    contributors may be used to endorse or promote products derived from
//    this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 

#pragma once

#include "CoreChemistryAPI.h"
#include "XMLParser.h"

// Defined in this header file:
class CDXMLElementHandler;
class CDXMLParser;

// Used in this file
class CDXDocument;

// *************************
// class XMLObjectIDTranslator
//
// this is a singleton, used to hold the translation table for cdxml element names
// *************************
class CDXMLObjectIDTranslator : public XMLObjectIDTranslator
{
public:
	CDXMLObjectIDTranslator();
	static CDXMLObjectIDTranslator* GetTranslator();
};

class CORE_CHEMISTRY_API CDXMLParser : public XMLParser
{
protected:
	virtual void startElement(const XML_Char *name, const XML_Char **atts);

	virtual CDXTag getAttributeID(const char *s) const;

	virtual bool IgnoreObject(CDXTag objTag, const CDXObject *parent) const { return false; }

public:
	CDXMLParser();

	virtual ~CDXMLParser();

	// Note: GetDocument does not relinquish ownership of the document, while ReleaseDocument does.
	CDXDocument *GetDocument();
	std::unique_ptr<CDXDocument> ReleaseDocument();
};
